@@ -1,3 +1,19 @@
+"""
+The helpers module contains utility functions that are used throughout the SDK and can also be used independently.
+
+---
+
+## Usage
+You can import the functions from the helpers module as follows:
+
+```python
+from bits_aviso_python_sdk.helpers import initialize_logger, convert_xml_to_dict, resolve_dns
+```
+
+---
+"""
+
+import base64
 import datetime
 import dns.resolver
 import json
@@ -5,18 +21,48 @@ import logging
 import os
 import re
 import xmltodict
+from dns.exception import Timeout
 from urllib.parse import urlparse
 
 
+def auth_basic(username, password):
+    """
+    Encode the username and password in base64 for HTTP Basic Authentication.
+
+    Args:
+        username (str): The username to authenticate with.
+        password (str): The password to authenticate with.
+
+    Returns:
+        str: The base64-encoded authentication token.
+    """
+    return base64.b64encode(f"{username}:{password}".encode("utf-8")).decode("ascii")
+
+
+def check_dict_keys(dict_to_check, keys):
+    """
+    Check if all specified keys are present in the given dictionary.
+
+    Args:
+        dict_to_check (dict): The dictionary to check.
+        keys (list): The keys to check for.
+
+    Returns:
+        bool: True if all keys are present, False otherwise.
+    """
+    return all(key in dict_to_check for key in keys)
+
+
 def convert_xml_to_dict(xml_string, json_output=False):
-    """Converts an XML string to a dictionary.
+    """
+    Convert an XML string to a dictionary or JSON object.
 
     Args:
         xml_string (str): The XML string to convert.
-        json_output (bool, optional): Whether to output the dictionary as a JSON. Defaults to False.
+        json_output (bool, optional): If True, return a JSON string. If False, return a dict. Defaults to False.
 
     Returns:
-        dict: The XML string converted to a dictionary.
+        dict or str: The XML string converted to a dictionary or JSON string.
     """
     if json_output:
         logging.debug("Converting XML to JSON object...")
@@ -28,12 +74,15 @@ def convert_xml_to_dict(xml_string, json_output=False):
 
 
 def initialize_logger(logger_level="INFO", file_handler_path=None):
-    """Initializes a logger with a stream handler and an optional file handler.
+    """
+    Initialize a logger with a stream handler and an optional file handler.
 
     Args:
-        logger_level (str, optional): The level of the logger. Defaults to "INFO".
-                                      Can be "DEBUG", "INFO", "WARNING", "ERROR", or "CRITICAL".
-        file_handler_path (str, optional): The path to save the log file if a file handler is desired. Defaults to None.
+        logger_level (str, optional): Logging level ("DEBUG", "INFO", "WARNING", "ERROR", or "CRITICAL"). Defaults to "INFO".
+        file_handler_path (str, optional): Directory path to save the log file. If None, no file handler is added. Defaults to None.
+
+    Returns:
+        logging.Logger: The configured logger instance.
     """
     # set up logger
     today = datetime.datetime.now().strftime("%Y_%m_%d")
@@ -87,13 +136,14 @@ def initialize_logger(logger_level="INFO", file_handler_path=None):
 
 
 def is_ip_address(string_to_check):
-    """Check if the given string is a valid IP address.
+    """
+    Check if the given string is a valid IPv4 address.
 
     Args:
         string_to_check (str): The string to check.
 
     Returns:
-        bool: True if the string is a valid IP address, False otherwise.
+        bool: True if the string is a valid IPv4 address, False otherwise.
     """
     ip_pattern = re.compile(r'^(\d{1,3}\.){3}\d{1,3}$')
     logging.debug(f"IP Pattern: {ip_pattern}")
@@ -107,19 +157,19 @@ def is_ip_address(string_to_check):
 
 
 def resolve_dns(domain, timeout=5, lifetime=15, dns_server=None, first_result_only=True, tcp=True):
-    """Resolves the domain to an IP address. If no DNS server is provided, it will use the system's default DNS server.
-    Timeout and lifetime are set to 5 and 15 seconds respectively by default.
+    """
+    Resolve a domain to one or more IP addresses using DNS.
 
     Args:
-        domain (str): The domain to resolve.
-        timeout (int, optional): The timeout for the DNS resolver. Defaults to 5.
-        lifetime (int, optional): The lifetime for the DNS resolver. Defaults to 10.
-        dns_server (str, optional): The DNS server to use. Defaults to None.
-        first_result_only (bool, optional): Whether to return only the first result. Defaults to True.
+        domain (str): The domain or URL to resolve.
+        timeout (int, optional): Timeout for the DNS resolver in seconds. Defaults to 5.
+        lifetime (int, optional): Lifetime for the DNS resolver in seconds. Defaults to 15.
+        dns_server (str, optional): DNS server to use. If None, uses system default. Defaults to None.
+        first_result_only (bool, optional): If True, return only the first result. If False, return all valid results. Defaults to True.
         tcp (bool, optional): Whether to use TCP for the query. Defaults to True.
 
     Returns:
-        str or list: The resolved IP address(es).
+        str or list: The resolved IP address(es). Returns a string if first_result_only is True, otherwise a list of strings.
     """
     # Check if the domain is a URL and extract the hostname
     if domain.startswith("http"):
@@ -144,7 +194,11 @@ def resolve_dns(domain, timeout=5, lifetime=15, dns_server=None, first_result_on
 
         # return the first result only if specified
         if first_result_only:
-            first_result = answer[0].to_text()
+            # Ensure answer is iterable and has at least one result
+            first_rdata = next(iter(answer), None)
+            if first_rdata is None:
+                raise ValueError(f"No DNS results for {domain}")
+            first_result = str(first_rdata)
             # check if the result is an IP address
             if not is_ip_address(first_result):
                 raise ValueError(f"Resolved {domain} to {first_result} which is not an IP address.")
@@ -169,7 +223,6 @@ def resolve_dns(domain, timeout=5, lifetime=15, dns_server=None, first_result_on
             return valid_ips
         else:
             raise ValueError(f"Unable to resolve {domain} to any valid IP addresses.")
-
-    except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN, dns.exception.Timeout, ValueError) as e:
+    except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN, Timeout, ValueError) as e:
         logging.error(f"Error resolving DNS: {e}")
         return
