@@ -3,8 +3,9 @@ This module provides a class to interact with Google Cloud Storage.
 If no service account credentials are provided, the SDK will attempt to use the default credentials.
 
 ## Usage
-The following is an example of how to upload a file to Google Cloud Storage:
+The following are examples of how to use the Storage class:
 
+### Upload a file to Google Cloud Storage
 ```python
 from bits_aviso_python_sdk.services.google.storage import Storage
 
@@ -18,33 +19,50 @@ file_to_upload = "path/to/your/file.txt"
 storage_client.upload("your_bucket_name", "prefix", "file.txt", file_to_upload)
 ```
 
+### Get the total size of a bucket (in bytes)
+```python
+from bits_aviso_python_sdk.services.google.storage import Storage
+
+# initialize Storage client
+storage_client = Storage()
+
+# get the total size of a bucket in bytes
+bucket_size_bytes = storage_client.get_bucket_size("your_bucket_name")
+print(f"Bucket size: {bucket_size_bytes} bytes")
+```
+
+The `get_bucket_size` method uses Google Cloud Monitoring (MQL via REST API) to fetch the latest total size of the specified bucket in bytes. It returns 0 if no data is found or if there is an error. All errors are logged and raised as `ValueError`.
+
 ---
 """
-import json
-
+import google.auth
 import google.auth.exceptions
 import io
+import json
 import logging
+import requests
 from google.api_core import exceptions, retry
 from google.cloud import storage, exceptions
+from google.auth.transport.requests import Request
 from bits_aviso_python_sdk.helpers.bigquery import parse_to_nldjson
 from bits_aviso_python_sdk.services.google import authenticate_google_service_account
 
 
 class Storage:
     def __init__(self, service_account_credentials=None, project_id=None):
-        """Initializes the Storage class. If service account credentials are not provided,
-        the credentials will be inferred from the environment.
+        """
+        Initialize the Storage class for interacting with Google Cloud Storage.
 
         Args:
-            service_account_credentials (dict, str, optional): The service account credentials in json format
-            or the path to the credentials file. Defaults to None.
-            project_id (str, optional): The project id to use. Defaults to None.
+            service_account_credentials (dict or str, optional): Service account credentials as a dict or path to the credentials file. If not provided, uses default credentials.
+            project_id (str, optional): The project ID to use. If not provided, uses the default project.
         """
         if service_account_credentials:
-            credentials = authenticate_google_service_account(service_account_credentials)
+            credentials = authenticate_google_service_account(
+                service_account_credentials)
             if project_id:
-                self.client = storage.Client(credentials=credentials, project=project_id)
+                self.client = storage.Client(
+                    credentials=credentials, project=project_id)
             else:
                 self.client = storage.Client(credentials=credentials)
         else:
@@ -58,16 +76,17 @@ class Storage:
                 self.client = None
 
     def download_blob_to_file(self, bucket_name, blob_name, file_path, prefix=None):
-        """Downloads the specified blob to a file.
+        """
+        Download a blob from a bucket to a local file.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            blob_name (string): The name of the blob.
-            file_path (string): The path to save the downloaded file.
-            prefix (string, optional): The prefix to use for the blob.
+            bucket_name (str): The name of the bucket.
+            blob_name (str): The name of the blob.
+            file_path (str): The path to save the downloaded file.
+            prefix (str, optional): The prefix (folder) for the blob. Defaults to None.
 
         Returns:
-            string: The path to the downloaded file.
+            str: The path to the downloaded file.
 
         Raises:
             ValueError: If the blob is not found in the bucket.
@@ -76,9 +95,11 @@ class Storage:
             # get the blob
             blob = self.get_blob(bucket_name, blob_name, prefix)
             # download the blob to the file
-            logging.info(f"Downloading [{blob_name}] from {bucket_name} to [{file_path}]...")
+            logging.info(
+                f"Downloading [{blob_name}] from {bucket_name} to [{file_path}]...")
             blob.download_to_filename(file_path)
-            logging.info(f"Downloaded [{blob_name}] from {bucket_name} to [{file_path}].")
+            logging.info(
+                f"Downloaded [{blob_name}] from {bucket_name} to [{file_path}].")
 
             return file_path
 
@@ -105,9 +126,11 @@ class Storage:
         """
         try:
             # create the blob
-            logging.info(f"Creating blob {prefix}/{blob_name} in bucket {bucket}...")
+            logging.info(
+                f"Creating blob {prefix}/{blob_name} in bucket {bucket}...")
             blob = bucket.blob(f"{prefix}/{blob_name}")
-            logging.info(f"Created blob {prefix}/{blob_name} in bucket {bucket}.")
+            logging.info(
+                f"Created blob {prefix}/{blob_name} in bucket {bucket}.")
 
             return blob  # return the blob
 
@@ -118,17 +141,16 @@ class Storage:
             raise ValueError(message)
 
     def get_blob(self, bucket_name, blob_name, prefix=None):
-        """Gets the specified blob. The blob_name refers to the equivalent of a "file" in a bucket.
-        The prefix is used to specify the "folder" in which the blob is located.
+        """
+        Retrieve a blob object from a bucket.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            blob_name (string): The name of the blob.
-            prefix (string, optional): The prefix to use for the blob.
-
+            bucket_name (str): The name of the bucket.
+            blob_name (str): The name of the blob (file).
+            prefix (str, optional): The prefix (folder) for the blob. Defaults to None.
 
         Returns:
-            google.cloud.storage.blob.Blob: The specified blob.
+            google.cloud.storage.blob.Blob: The blob object.
 
         Raises:
             ValueError: If the blob is not found in the bucket.
@@ -156,15 +178,16 @@ class Storage:
             raise ValueError(message)
 
     def get_blob_dict(self, bucket_name, blob_name, prefix=None):
-        """Gets the information for the specified blob and returns it as a dictionary.
+        """
+        Get metadata for a blob as a dictionary.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            blob_name (string): The name of the blob.
-            prefix (string, optional): The prefix to use for the blob.
+            bucket_name (str): The name of the bucket.
+            blob_name (str): The name of the blob.
+            prefix (str, optional): The prefix (folder) for the blob. Defaults to None.
 
         Returns:
-            dict: The metadata for the specified blob.
+            dict: Metadata for the specified blob.
         """
         # get the blob
         blob = self.get_blob(bucket_name, blob_name, prefix)
@@ -210,17 +233,22 @@ class Storage:
         return blob_data
 
     def get_bucket(self, bucket_name):
-        """Gets the specified bucket.
+        """
+        Retrieve a bucket object by name.
 
         Args:
-            bucket_name (string): The name of the bucket.
+            bucket_name (str): The name of the bucket.
 
         Returns:
-            google.cloud.storage.bucket.Bucket: The specified bucket.
+            google.cloud.storage.bucket.Bucket: The bucket object.
 
         Raises:
             ValueError: If the bucket is not found.
         """
+        if not self.client:
+            message = "Storage client is not initialized."
+            logging.error(message)
+            raise ValueError(message)
         try:
             # get_bucket the bucket
             logging.info(f"Retrieving bucket {bucket_name}...")
@@ -236,14 +264,15 @@ class Storage:
             raise ValueError(message)
 
     def get_bucket_dict(self, bucket_name=None, bucket_obj=None):
-        """Gets the data for the specified bucket based on the bucket name or object and returns it as a dictionary.
+        """
+        Get metadata for a bucket as a dictionary.
 
         Args:
-            bucket_name (string, optional): The name of the bucket. Defaults to None.
+            bucket_name (str, optional): The name of the bucket. Defaults to None.
             bucket_obj (google.cloud.storage.bucket.Bucket, optional): The bucket object. Defaults to None.
 
         Returns:
-            dict: The metadata for the specified bucket.
+            dict: Metadata for the specified bucket.
         """
         # check if the bucket object is provided
         if bucket_obj:
@@ -289,98 +318,128 @@ class Storage:
         return bucket_data
 
     def get_bucket_size(self, bucket):
-        """Gets the total size of the specified bucket.
+        """
+        Get the total size of a bucket in bytes using Cloud Monitoring (MQL via REST API).
 
         Args:
-            bucket (string or google.cloud.storage.bucket.Bucket): The name of the bucket or the bucket object.
+            bucket (str or google.cloud.storage.bucket.Bucket): The bucket name as a string, or a Bucket object.
 
         Returns:
-            int: The total size of the bucket in bytes.
+            int: The total size of the bucket in bytes. Returns 0 if no data is found.
 
         Raises:
-            ValueError: If the bucket is not found.
+            ValueError: If the storage client is not initialized, or if there is an error querying or parsing the Monitoring API response.
         """
+        if isinstance(bucket, str):
+            bucket_name = bucket
+        else:
+            bucket_name = bucket.name
+
+        if not self.client or not hasattr(self.client, "project"):
+            raise ValueError("Storage client is not initialized or missing project attribute.")
+        project = self.client.project
+
+        # Get credentials and access token
+        credentials, _ = google.auth.default(scopes=[
+            "https://www.googleapis.com/auth/cloud-platform"
+        ])
+        credentials.refresh(Request())
+        access_token = credentials.token
+
+        url = f"https://monitoring.googleapis.com/v3/projects/{project}/timeSeries:query"
+
+        mql_query = f"""
+            fetch gcs_bucket
+            | metric 'storage.googleapis.com/storage/total_bytes'
+            | filter resource.bucket_name == '{bucket_name}'
+            | within 1d
+            | group_by [], [max_total_bytes: max(value.total_bytes)]
+        """
+        headers = {
+            "Authorization": f"Bearer {access_token}",
+            "Content-Type": "application/json"
+        }
+        body = {"query": mql_query}
+
         try:
-            # get the bucket
-            if isinstance(bucket, str):
-                bucket = self.get_bucket(bucket)
+            response = requests.post(url, headers=headers, json=body)
+            response.raise_for_status()
 
-            # variables
-            total_size = 0
+        except requests.HTTPError as http_err:
+            logging.error(f"HTTP error when querying Monitoring API for bucket {bucket_name}: {http_err} | Response: {getattr(http_err.response, 'text', None)}")
+            raise ValueError(f"Failed to fetch bucket size for {bucket_name}: {http_err}")
+        except Exception as req_err:
+            logging.error(f"Unexpected error when querying Monitoring API for bucket {bucket_name}: {req_err}")
+            raise ValueError(f"Failed to fetch bucket size for {bucket_name}: {req_err}")
 
-            # iterate through all objects in the bucket and sum their sizes
-            logging.info(f"Retrieving size for bucket {bucket.name}...")
-            for blob in bucket.list_blobs():
-                total_size += blob.size
+        try:
+            data = response.json()
 
-            logging.info(f"Successfully retrieved size for bucket {bucket.name}. Total Size: {total_size} bytes.")
-            return total_size
+        except Exception as json_err:
+            logging.error(f"Failed to parse Monitoring API response for bucket {bucket_name}: {json_err}")
+            raise ValueError(f"Failed to parse Monitoring API response for {bucket_name}: {json_err}")
 
-        except exceptions.NotFound:
-            message = f"Bucket {bucket} not found."
-            logging.error(message)
-            raise ValueError(message)
+        try:
+            latest = None
+            latest_time = None
+            for ts in data.get("timeSeriesData", []):
+                for point in ts.get("pointData", []):
+                    values = point.get("values", [])
+                    if values:
+                        v = values[0]
+                        val = v.get("int64Value")
+                        if val is None:
+                            val = v.get("doubleValue")
+                        t = point.get("timeInterval", {}).get("endTime")
+                        if val is not None and (latest_time is None or t > latest_time):
+                            latest = val
+                            latest_time = t
+            if latest is not None:
+                return int(float(latest))
+            logging.warning(f"No bucket size data found for {bucket_name} in Monitoring API response.")
+            return 0
 
-        except exceptions.GoogleCloudError as e:
-            message = f"Error retrieving size for given bucket argument [{bucket}]: {e}"
-            logging.error(message)
-            raise ValueError(message)
+        except Exception as parse_err:
+            logging.error(f"Error parsing MQL response for bucket {bucket_name}: {parse_err}")
+            raise ValueError(f"Failed to parse bucket size for {bucket_name}: {parse_err}")
 
     def list_blobs(self, bucket_name, prefix=None, delimiter=None):
-        """Lists all the blobs in the bucket that begin with the prefix.
-
-        This can be used to list all blobs in a "folder", e.g. "public/".
-
-        The delimiter argument can be used to restrict the results to only the
-        "files" in the given "folder". Without the delimiter, the entire tree under
-        the prefix is returned. For example, given these blobs:
-
-            a/1.txt
-            a/b/2.txt
-
-        If you specify prefix ='a/', without a delimiter, you'll get back:
-
-            a/1.txt
-            a/b/2.txt
-
-        However, if you specify prefix='a/' and delimiter='/', you'll get back
-        only the file directly under 'a/':
-
-            a/1.txt
-
-        As part of the response, you'll also get back a blobs.prefixes entity
-        that lists the "subfolders" under `a/`:
-
-            a/b/
-
-        Copied from Google Cloud Storage documentation.
+        """
+        List all blobs in a bucket, optionally filtering by prefix and delimiter.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            prefix (string, optional): The prefix to use for the blob. Defaults to None.
-            delimiter (string, optional): The delimiter to use to restrict the results. Defaults to None.
+            bucket_name (str): The name of the bucket.
+            prefix (str, optional): The prefix (folder) to filter blobs. Defaults to None.
+            delimiter (str, optional): The delimiter to restrict results. Defaults to None.
 
         Returns:
-            list: A list of all the blobs in google object form in the bucket.
+            list: List of blob objects in the bucket.
         """
         # list the blobs in the bucket
-        logging.info(f"Listing blobs in [{bucket_name}] with prefix [{prefix}] and delimeter [{delimiter}]...")
-        return [b for b in self.client.list_blobs(bucket_name, prefix=prefix, delimiter=delimiter)]
+        logging.info(
+            f"Listing blobs in [{bucket_name}] with prefix [{prefix}] and delimeter [{delimiter}]...")
+        if not self.client:
+            message = "Storage client is not initialized."
+            logging.error(message)
+            raise ValueError(message)
+        return list(self.client.list_blobs(bucket_name, prefix=prefix, delimiter=delimiter))
 
     def list_blobs_dict(self, bucket_name, prefix=None, delimiter=None):
-        """Lists all the blobs in the specified bucket. If a prefix is provided, only the blobs with the prefix will be
-        listed. If a delimiter is provided, the results will be restricted to the specified delimiter.
+        """
+        List all blobs in a bucket and return their metadata as dictionaries.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            prefix (string, optional): The prefix to use for the blob. Defaults to None.
-            delimiter (string, optional): The delimiter to use to restrict the results. Defaults to None.
+            bucket_name (str): The name of the bucket.
+            prefix (str, optional): The prefix (folder) to filter blobs. Defaults to None.
+            delimiter (str, optional): The delimiter to restrict results. Defaults to None.
 
         Returns:
-            list: A list of dictionaries representing the blobs in the bucket.
+            list: List of dictionaries representing the blobs in the bucket.
         """
-        logging.info(f"Getting blob dict data in [{bucket_name}] with prefix [{prefix}] and delimeter [{delimiter}]...")
-        blobs = self.list_blobs(bucket_name, prefix=prefix, delimiter=delimiter)
+        logging.info(
+            f"Getting blob dict data in [{bucket_name}] with prefix [{prefix}] and delimeter [{delimiter}]...")
+        blobs = self.list_blobs(
+            bucket_name, prefix=prefix, delimiter=delimiter)
         blobs_dict = []
         for b in blobs:
             blobs_dict.append(self.get_blob_dict(bucket_name, b.name, prefix))
@@ -389,21 +448,31 @@ class Storage:
         return blobs_dict
 
     def list_buckets(self):
-        """Lists all the buckets in the project. Each item in the list is a google bucket object.
+        """
+        List all buckets in the project.
 
         Returns:
-            list: A list of all the buckets in the project.
+            list: List of bucket objects in the project.
         """
         logging.info("Getting all bucket objects...")
-        return [b for b in self.client.list_buckets()]
+        if not self.client:
+            message = "Storage client is not initialized."
+            logging.error(message)
+            raise ValueError(message)
+        return list(self.client.list_buckets())
 
     def list_buckets_dict(self):
-        """Lists all the buckets in the project. Each item in the list is a dictionary representing the bucket.
+        """
+        List all buckets in the project and return their metadata as dictionaries.
 
         Returns:
-            list: The list of buckets in the project.
+            list: List of dictionaries representing the buckets in the project.
         """
         logging.info("Getting all bucket dict data...")
+        if not self.client:
+            message = "Storage client is not initialized."
+            logging.error(message)
+            raise ValueError(message)
         buckets = self.client.list_buckets()
         buckets_dict = []
         for b in buckets:
@@ -412,19 +481,80 @@ class Storage:
         logging.info("Finished getting all bucket dict data.")
         return buckets_dict
 
-    def upload(self, bucket_name, prefix, blob_name, data, content_type='application/json', nldjson=False):
-        """Uploads the data to the specified bucket. The data must be a string, dictionary, or list.
+    def update_bucket_labels(self, bucket_name, labels, append=True):
+        """
+        Update the labels for a bucket.
 
         Args:
-            bucket_name (string): The name of the bucket.
-            prefix (string): The prefix to use for the blob. Typically, the name of the dataset folder.
-            blob_name (string): The name of the blob.
-            data (str, dict, list): The data to be uploaded to the bucket.
-            content_type (string, optional): The content type of the data. Defaults to 'application/json'.
-            nldjson (bool, optional): Whether to convert data to newline delimited json. Defaults to False.
+            bucket_name (str): The name of the bucket.
+            labels (dict): The labels to update the bucket with.
+            append (bool, optional): If True, append to existing labels; if False, replace them. Defaults to True.
+
+        Returns:
+            google.cloud.storage.bucket.Bucket: The updated bucket object.
 
         Raises:
-            TypeError: If the data cannot be converted to newline delimited json.
+            ValueError: If the bucket is not found or if there is an error updating the labels.
+        """
+        try:
+            # get the bucket
+            bucket = self.get_bucket(bucket_name)
+            # check for existing labels
+            existing_labels = bucket.labels or {}
+
+            # check if the labels are valid
+            if not isinstance(labels, dict):
+                message = "Labels must be a dictionary."
+                logging.error(message)
+                raise ValueError(message)
+
+            # check if the labels are empty
+            if not labels:
+                message = "No labels provided to update."
+                logging.error(message)
+                raise ValueError(message)
+
+            # check whether to append or replace the labels
+            if append:
+                # append the new labels to the existing labels
+                logging.info(
+                    f"Appending new labels to existing labels for bucket {bucket_name}...")
+                labels = {**existing_labels, **labels}
+
+            # update the labels
+            logging.info(f"Updating labels for bucket {bucket_name}...")
+            bucket.labels = labels
+            bucket.patch()  # apply the changes
+            logging.info(f"Updated labels for bucket {bucket_name}.")
+
+            return bucket
+
+        except exceptions.NotFound:
+            message = f"Bucket {bucket_name} not found."
+            logging.error(message)
+
+            raise ValueError(message)
+
+        except exceptions.GoogleCloudError as e:
+            message = f"Error updating labels for bucket {bucket_name}: {e}"
+            logging.error(message)
+
+            raise ValueError(message)
+
+    def upload(self, bucket_name, prefix, blob_name, data, content_type='application/json', nldjson=False):
+        """
+        Upload data to a bucket as a blob.
+
+        Args:
+            bucket_name (str): The name of the bucket.
+            prefix (str): The prefix (folder) for the blob.
+            blob_name (str): The name of the blob.
+            data (str, dict, list): The data to upload.
+            content_type (str, optional): The content type of the data. Defaults to 'application/json'.
+            nldjson (bool, optional): If True, convert data to newline-delimited JSON. Defaults to False.
+
+        Raises:
+            TypeError: If the data cannot be converted to newline-delimited JSON.
             ValueError: If the data cannot be uploaded to the bucket.
         """
         try:
@@ -443,7 +573,8 @@ class Storage:
                     data = parse_to_nldjson(data)
 
                 except TypeError as e:  # data is not a dictionary or a list of dictionaries, probably already converted
-                    raise ValueError(f"Unable to convert data to newline delimited json. {e}")
+                    raise ValueError(
+                        f"Unable to convert data to newline delimited json. {e}")
 
             # check if the data is a string, if not convert it to string
             if isinstance(data, dict) or isinstance(data, list):
@@ -454,7 +585,8 @@ class Storage:
 
             # upload the data
             logging.info(f"Uploading {prefix}/{blob_name} to {bucket_name}...")
-            blob.upload_from_file(stream, retry=retry_policy, content_type=content_type, timeout=120)
+            blob.upload_from_file(stream, retry=retry_policy,
+                                  content_type=content_type, timeout=120)
             logging.info(f"Uploaded {prefix}/{blob_name} to {bucket_name}.")
 
         except (ValueError, AttributeError) as e:
