@@ -11,8 +11,8 @@ def test():
     clusters_secret = ''
     credentials = secret_manager.get_secret(google_project, credentials_secret)
     clusters = secret_manager.get_secret(google_project, clusters_secret)
-    username = credentials.get('username')
-    password = credentials.get('password')
+    username = credentials.get('username')  # type: ignore
+    password = credentials.get('password')  # type: ignore
     isilon = Isilon(username, password, clusters)
     quota_data, _ = isilon.get_entity_for_all_clusters('quotas', '/quota/quotas')
     write_json(quota_data, 'quotas.json')
